#!/usr/bin/env python
"""
vassal_config

Object that represents a uwsgi vassal configuration with some
extra fields that can be used to control setup for the application
including creation of a virtualenv and pip installing requirements

vassal configs containing an extra section called [vassaldeployer]
as well as the initial [uwsgi] config get processed and included
in a generated nginx config file that contains the site definition
for the multiple uwsgi vassal apps

Eg:

[uwsgi]
home=/opt/app
socket=127.0.0.1:3030
module=some_package.some_module:APP
master=1
enable-threads=true
workers=2
die-on-term=1
virtualenv=/opt/app/venv

[vassaldeployer]
app_url=/some_package
python=python2.7
requirements=some_package=0.1.2.some_dep==1.2.3
pip_options= --extra-index=mypypi:8080

This script will create the virtualenv in the location specified,
install the requirements via pip and add a section to
the master nginx conf mapping /some_package to the uwsgi app
listening on the specified socket

"""
import os
from virtualenvapi.manage import VirtualEnvironment
from . import logger


try:
    import ConfigParser as configparser
except ImportError:  # pragma: no cover
    import configparser


class VassalConfig(dict):
    """
    Representation of a uwsgi vassal config, taking a
    templated uwsgi vassal config with additional directives controlled
    by a vassaldeployer section
    """
    SECTION = 'vassaldeployer'

    def __init__(self, conf_file):
        self.config_file = conf_file
        self.basename = os.path.basename(self.config_file)

    def load(self):
        """
        read the ini file
        """
        self.parser = configparser.RawConfigParser()
        self.parser.read(self.config_file)
        for section in self.parser.sections():
            self.setdefault(section, {})
            for option in self.parser.options(section):
                self[section].setdefault(
                    option,
                    self.parser.get(section, option)
                )

    def uwsgi_config(self, **settings):
        """
        write the cleaned uwsgi config including
        additional settings
        """
        content = "[uwsgi]\n"
        values = dict(self['uwsgi'])
        values.update(settings)
        content += '\n'.join("{}={}".format(k, v) for k, v in iter(values.items()))
        return content

    def nginx_config(self):
        """
        make the location section of the nginx conf

        routing the apps based off a subdomain of the nginx
        server requires the two extra params:
          SCRIPT_NAME tells nginx to strip the url
          uwsgi_modifier1 30 is the magic that makes it work
          see: http://blog.codepainters.com/2012/08/05/wsgi-deployment-under-a-subpath-using-uwsgi-and-nginx/
          for more details.

        """
        conf = "location {}".format(self.app_url)
        conf += "{\n"
        conf += "    include uwsgi_params;\n"
        conf += "    uwsgi_pass 127.0.0.1:{};\n".format(self.uwsgi_port)
        conf += "    uwsgi_param SCRIPT_NAME {};\n".format(self.app_url)
        conf += "    uwsgi_modifier1 30;\n"
        conf += "}\n"
        return conf

    @property
    def uwsgi_socket(self):
        """uwsgi socket location, linked to nginx"""
        return self['uwsgi']['socket']

    @property
    def uwsgi_home(self):
        """uwsgi home param"""
        return self['uwsgi']['home']

    @property
    def uwsgi_virtualenv(self):
        """virtualenv location"""
        if not self['uwsgi'].get('virtualenv'):
            venv = os.path.join(self.uwsgi_home, 'venv')
            self['uwsgi']['virtualenv'] = venv
        return self['uwsgi']['virtualenv']

    @property
    def uwsgi_port(self):
        """uwsgi port, linked to nginx"""
        return int(self.uwsgi_socket.split(':', 1)[1])

    @property
    def section(self):
        """get the template section"""
        return self[self.SECTION]

    @property
    def app_url(self):
        """app url setting"""
        return self.section.get('app_url')

    @property
    def app_python(self):
        """app python interpreter"""
        return self.section.get('python', 'python')

    @property
    def app_requirements(self):
        """app requirements to install"""
        return self.section.get('requirements', '')

    @property
    def pip_options(self):
        """extra options to pass to pip install"""
        return [x.strip() for x in self.section.get('pip_options', '').split() if x.strip()]

    def make_virtualenv(self):
        """
        execute the virtualenv command using the appropriate python
        """
        logger.info("Making venv: {}".format(self.uwsgi_virtualenv))
        if not os.path.exists(self.uwsgi_home):
            os.makedirs(self.uwsgi_home)
        env = VirtualEnvironment(self.uwsgi_virtualenv, python=self.app_python)
        env.open_or_create()

    def pip_install(self):
        """
        pip install the requirements list
        """
        reqs = [
            x.strip()
            for x in self.app_requirements.split(',') if x.strip()
        ]
        if not reqs:
            # no reqs specified
            return
        env = VirtualEnvironment(self.uwsgi_virtualenv)
        for req in reqs:
            logger.info("installing {} in {}".format(req, self.uwsgi_virtualenv))
            env.install(req, options=self.pip_options)

    def write(self, dirname):
        """write config for vassal into the vassals dir"""
        f = os.path.join(dirname, self.basename)
        logger.info("writing vassal config: {}".format(f))
        with open(f, 'w') as handle:
            handle.write(self.uwsgi_config())



#!/usr/bin/env python
"""
deploy

"""
import os
from .vassal_config import VassalConfig
from .nginx_config import NginxSite
from . import logger


def list_vassals_configs(directory):
    """
    find all *.ini files in the given directory path

    :param directory: file system path containing *.ini files

    :returns: list of abs path ini files found
    """
    result = []
    for f in os.listdir(directory):
        if f.endswith('.ini'):
            result.append(
                os.path.join(directory, f)
            )
    return result


def make_vassals(directory):
    """
    Find the *.ini files in the given directory and wrap each
    one in a VassalConfig object

    :param directory: file system path containing *.ini files
    :returns: list of VassalConfig instances, one per found file

    """
    vassals = []
    for vc in list_vassals_configs(directory):
        logger.info('found vassal config: {}'.format(vc))
        vassals.append(VassalConfig(vc))
    return vassals


def deploy(
        templates_dir,
        vassals_dir,
        site_name,
        site_port,
        sites_available=None,
        sites_enabled=None):
    """
    _deploy_

    Main call to consume a directory containing the vassal templates,
    write the rendered vassal configs to the vassals_dir,
    set up the virtualenvs as needed and build the nginx site config

    :param templates_dir: file system path containing uwsgi *.ini files
    :param vassals_dir: file system path to write the vassal configs
        to for deployment
    :param site_name: nginx site name
    :param site_port: nginx port number
    :param sites_available: Path to nginx sites available dir
        (defaults to /etc/nginx/sites-available)
    :param sites_enabled: Path to nginx sites enabled dir
        (defaults to /etc/nginx/sites-enabled)

    """
    vassals = make_vassals(templates_dir)
    site = NginxSite(site_name, site_port, sites_available, sites_enabled)

    for vassal in vassals:
        vassal.load()
        vassal.make_virtualenv()
        vassal.pip_install()
        vassal.write(vassals_dir)
        site.add_vassal(vassal)

    site.write_available()
    site.link_enabled()



#!/usr/bin/env python
"""
vassal_deployer

"""
__version__="0.0.1"

import os

from .logger import get_logger

logger = get_logger(
    os.environ.get('VASSAL_DEPLOYER_LOG'),
    os.environ.get('VASSAL_DEPLOYER_STDOUT', False)
)



#!/usr/bin/env python
"""
cli entry point for vassal deployer

"""
import argparse
from .deploy import deploy


def build_parser():
    """
    build command line parser

    """
    parser = argparse.ArgumentParser(
        description='uwsgi vassal config processor that builds nginx confs'
    )
    parser.add_argument(
        '--vassals',
        help='directory to write vassals configs',
        required=True,
        dest='vassals_out'
    )
    parser.add_argument(
        '--input-vassals', '-i',
        help='directory containing vassals configs with extra deployer conf section',
        required=True,
        dest='vassals_in'
    )
    parser.add_argument(
        '--sites-enabled',
        help='nginx sites-enabled directory location',
        default='/etc/nginx/sites-enabled',
        dest='sites_enabled'
    )
    parser.add_argument(
        '--sites-available',
        help='nginx sites-available directory location',
        default='/etc/nginx/sites-available',
        dest='sites_available'
    )
    parser.add_argument(
        '--nginx-port',
        help='nginx server port number',
        default=8080,
        dest='nginx_port'
        )
    parser.add_argument(
        '--nginx-site',
        help='nginx site name',
        default='uwsgi_vassals',
        dest='nginx_site'
        )

    opts = parser.parse_args()
    return opts


def main():
    """
    parse cli args and run deploy
    """
    opts = build_parser()
    deploy(
        opts.vassals_in,
        opts.vassals_out,
        opts.nginx_site,
        opts.nginx_port,
        opts.sites_available,
        opts.sites_enabled
        )

if __name__ == '__main__':  # pragma: no cover
    main()




#!/usr/bin/env python
"""
logger setup
"""
import sys
import logging


_LOGGERS = {
    'LOGGER': None,
    'FILE_HANDLER': None,
    'STDOUT_HANDLER': None
}


STDOUT_FORMATTER = logging.Formatter(
    "%(asctime)s;%(message)s",
    datefmt="%Y-%m-%d %H:%M:%S"
)

FILE_FORMATTER = logging.Formatter(
    "%(asctime)s;%(levelname)s;%(message)s",
    datefmt="%Y-%m-%d %H:%M:%S"
)


def get_logger(logfile=None, stdout=True):
    if _LOGGERS['LOGGER']:
        return _LOGGERS['LOGGER']

    log = logging.getLogger('vassal_deployer')
    log.setLevel(logging.DEBUG)
    if stdout and (_LOGGERS['STDOUT_HANDLER'] is None):
        so_handler = logging.StreamHandler(stream=sys.stdout)
        so_handler.setLevel(logging.DEBUG)
        so_handler.setFormatter(STDOUT_FORMATTER)
        _LOGGERS['STDOUT_HANDLER'] = so_handler
        if so_handler not in log.handlers:
            log.addHandler(so_handler)

    if logfile and (_LOGGERS['FILE_HANDLER'] is None):
        handler = logging.handlers.WatchedFileHandler(logfile)
        handler.setLevel(logging.DEBUG)
        handler.setFormatter(FILE_FORMATTER)
        _LOGGERS['FILE_HANDLER'] = handler
        if handler not in log.handlers:
            log.addHandler(handler)

    _LOGGERS['LOGGER'] = log
    return _LOGGERS['LOGGER']



#!/usr/bin/env python
"""
nginx_config

Wrapper to build an nginx site config from a set of
uwsgi vassal configurations

"""
import os
from . import logger


class NginxSite(object):

    """
    Helper to build the nginx config file for the
    uwsgi site containing the vassals.

    Instantiate with the site name and port and optional locations of
    the nginx sites available and sites enabled dirs.

    """

    def __init__(
            self,
            site_name,
            site_port,
            sites_available=None,
            sites_enabled=None):
        self.site_name = site_name
        self.site_port = site_port
        self.sites_available = sites_available or '/etc/nginx/sites-available'
        self.sites_enabled = sites_enabled or '/etc/nginx/sites-enabled'
        self._vassals = []
        self._conf_file = "{}.conf".format(site_name)

    @property
    def conf_header(self):
        """nginx config header with uwsgi params"""
        nginx_conf = (
            "server {{\n"
            "    listen {};\n"
            "    server_tokens off;\n"
            "    server_name {};\n\n"
        ).format(self.site_port, self.site_name)
        return nginx_conf

    @property
    def conf_footer(self):
        """config footer"""
        return "\n}\n"

    def add_vassal(self, vassal_conf):
        """
        add a uwsgi vassal to this config via a VassalConfig instance

        :param vassal_conf: Instance of VassalConfig to be included in the
           nginx config
        """
        self._vassals.append(vassal_conf)

    def configuration(self):
        """
        build the nginx configuration string
        """
        conf = "{}".format(self.conf_header)
        conf += "\n".join(v.nginx_config() for v in self._vassals)
        conf += self.conf_footer
        return conf

    def write_available(self):
        """
        write the nginx config to the sites available directory
        """
        avail_file = os.path.join(self.sites_available, self._conf_file)
        logger.info("writing sites available: {}".format(avail_file))
        with open(avail_file, 'w') as handle:
            handle.write(self.configuration())

    def link_enabled(self):
        """
        link the config file for the site from sites-available to sites-enabled
        """
        avail_file = os.path.join(self.sites_available, self._conf_file)
        enabled_file = os.path.join(self.sites_enabled, self._conf_file)
        logger.info("linking {} to {}".format(avail_file, enabled_file))
        os.system("ln -sf {} {}".format(avail_file, enabled_file))
